'''

 Plug-in Title : "TexturizeByColourAreaSE"
 Gimp Menu     : "<Image>/Diego/Quantize" 
 Release       :  1.0    (initial)
 Author        :  Diego Nassetti (dinasset@gimpchat.com)
 Licence       :  GNU General Public License
 
 System Environment used in development 
 --------------------------------------                
 Platform      : Windows 10 64-bits
 Graphic Apps  : Gimp 2.8.16
 Language      : python as included in Gimp
                 
 Extrn Filters : (Gimp) script_fu_chisel (by Rob Antonishen)
                 available under Gimp-chat Scripts&Plug-ins as "Chisel Or Carve"                
 
 Summary
 -------
 Filter to implement the creation on a NEW IMAGE (dimension is a user option)
 starting from an RGB* input image
 where colours are quantized using the temporary conversion of the image
 from RGB to Indexed (nr of colours is a user option)
 each color area is filled with a SELECTED SET OF TEXTURES (user option).
 Filter has been created on the basis of Gimp-chat thread "Let's carve some wood".
 Even if the main use will be with patterns in wood, it has been generalized
 so that you can use any set of textures you like, patterns or gradients
 (for this reason "wood" does not appear in the filter's name)
 
 Details       
 -------
 Flow implemented to get the final result:      
 
 1. calculate the new image width and height scale for the output (user choice for dimensions)
 2. save and set-up context values
 3. create the new image by appropriately scaling the input image dimensions
 4. create the BG layer for the new image by copying the input layer, scaled
 5. add alpha to the layer is not present
 6. create a new layer (dup of the BG) for smoothed colours, SOFTLIGHT
 7. apply Gimp Selective Gaussina Blur
 8. create a new layer (dup of the Smoothed Layer) for the quantization 
 9. transform the image from RGBA to Indexed, with the user chosen number of colours
 10.create a temporary palette
 11.back to RGBA
 12.load the complete set of textures chosen: array of names and calculated luminosity values
    the textures may be patterns as well as gradients
 13.loop on each different color area, selected by using the palette
    a- if the user choice is for having the best texture matching by luma, do so
       otherwise choose a random texture
    b- if the colour area (in pixels) is small then the width of the effect is small and the effect is chisel
       if the colour area (in pixels) is big then the width of the effect is big and the effect is carve
       otherwise the width of the effect is medium and the effect is chisel
    c- fill the area with the chosen pattern/gradient
    d- apply the chisel/carve filter
    e- merge the layer created by the chisel into the source layer
 14.set the texturized layer to GRAIN-MERGE
 15.create a new layer from visible and set to SOFT-LIGHT at 45% opacity
 16.if the user chose to perform an artistic post-processing,
    apply the selected one
 17.removed
 18.if the user chose to move the background layer on top,
    raise BG to the top and even duplicate it to allow
    the use of combined modes and opacities
 19.Analyze the BG Layer luminosity and set the top layer mode to:
    - color if the image is very dark
    - multiply if the image is very light
    - hard-light otherwise   
 20.flatten the image is requested, recalling nr of colours and type of filler in the drawable name
 
 Acknowledgments
 ---------------
 This filter has been thoroughly tested by my dear gimp-chat friend Issabella (thanks a lot!)
 her suggestions were carefully taken into consideration for the final release
 
 
 ==================================================================================================
 
 ----------------------------------------------------------------
 COPYRIGHT NOTICE
 ----------------------------------------------------------------
 
 This program is free software you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation either version 3 of the License, or
 (at your option) any later version.  
 
 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with this program if not, you can view the GNU General Public
 License version 3 at the web site http://www.gnu.org/licenses/gpl-3.0.html
 Alternatively you can write to the Free Software Foundation, Inc., 675 Mass
 Ave, Cambridge, MA 02139, USA.
 
 -------------------------------------------------------------------
 other info
 -------------------------------------------------------------------
 
'''
#!/usr/bin/env python
from gimpfu import *
import decimal
import random


# service routines


def choose_appropriate_texture (areaColor, LFORM, LTEXT):
            
    closest="**NONE**"              # default
    distance=999999                 # limit
    (R, G, B, T)=areaColor 
    #areaf=R*R+G*G+B*B
    areaf=int(float(R*0.299)+float(G*0.587)+float(B*0.114))  #new, based on luminosity
    ##pdb.gimp_message("this area color formula is "+str(areaf))
    il=len(LFORM)
    for i in range(0, il):
        this_distance=abs(areaf-LFORM[i])
        ##pdb.gimp_message("the distance from "+LP[i]+" is "+str(this_distance))
        if this_distance<distance:
            distance=this_distance
            closest=LTEXT[i]
    textureName=closest
    #pdb.gimp_message("closest = "+textureName)
    
    return(textureName)

def load_patterns(names_filter):    # load all patterns
    
    np = 0  # nr of patterns loaded
    LPATT = []  # list for mixed patterns
    (np, LPATT) = pdb.gimp_patterns_get_list (names_filter)       
    
    tempImage = pdb.gimp_image_new(256, 256, RGB)
    tempLayer = pdb.gimp_layer_new(tempImage,256,256,RGB_IMAGE,"temp",100.0,NORMAL_MODE)
    tempImage.add_layer(tempLayer, 0)

    LFORM = []  # list for formula value
    
    if (np == 0):
        pdb.gimp_message("ERROR! Folder for patterns does not contain any valid pattern. Exit.")
        exit
    else :
        gimp.message ("nr of patterns = "+str(np))
        for i in range (np) :
        
            pdb.gimp_context_set_pattern (LPATT[i])
            pdb.gimp_drawable_fill (tempLayer, PATTERN_FILL)
    
            (meanR, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 1, 0, 255)
            (meanG, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 2, 0, 255)
            (meanB, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 3, 0, 255)
            aCformula=int(meanR)*int(meanR)+int(meanG)*int(meanG)+int(meanB)*int(meanB)
            aCformula=int(float(meanR*0.299)+float(meanG*0.587)+float(meanB*0.114))       #luminance=(red*0.299+green*0.587+blue*0.114)
           
            LFORM.append(aCformula)

    
    return(LPATT,LFORM)

def load_gradients(names_filter):    # load all gradients
    
    ng = 0  # nr of gradients loaded
    LGRAD = []  # list for mixed gradients
    (ng, LGRAD) = pdb.gimp_gradients_get_list (names_filter)       # "Mandel"
    
    tempImage = pdb.gimp_image_new(256, 1, RGB)
    tempLayer = pdb.gimp_layer_new(tempImage,256,1,RGB_IMAGE,"temp",100.0,NORMAL_MODE)
    tempImage.add_layer(tempLayer, 0)

    LFORM = []  # list for grad formula value
    
    if (ng == 0):
        pdb.gimp_message("ERROR! Folder for gradients does not contain any valid gradient. Exit.")
        exit
    else :
        gimp.message ("nr of gradients = "+str(ng))
        for i in range (ng) :
        
            pdb.gimp_context_set_gradient (LGRAD[i])
            pdb.gimp_edit_blend (tempLayer, CUSTOM_MODE, NORMAL_MODE, GRADIENT_LINEAR, 100.0,
                             0, REPEAT_NONE, False, False, 1, 0, False, 0, 0, 255, 0)
    
            (meanR, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 1, 0, 255)
            (meanG, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 2, 0, 255)
            (meanB, stddev, median, pixels, count, percent)=pdb.gimp_histogram(tempLayer, 3, 0, 255)
            aCformula=int(meanR)*int(meanR)+int(meanG)*int(meanG)+int(meanB)*int(meanB)
            aCformula=int(float(meanR*0.299)+float(meanG*0.587)+float(meanB*0.114))       #luminance=(red*0.299+green*0.587+blue*0.114)
           
            LFORM.append(aCformula)

    
    return(LGRAD,LFORM)



# ************ # 
# MAIN PROGRAM #
# ------------ #

def TexturizeByColourAreaSE(inImage, drawable,
                widthheight, 
                nr_of_colours,
                namesFilter,
                fillType,
                selMethod,
                ccRoundness,
                BGonTop,
                doFlatten):
  
   
    # image characteristics
    width = inImage.width
    height = inImage.height
    type = inImage.base_type
    # calculate scale factor for the new output image
    optiWidthHeight = widthheight       # size to work on to get best results
    realWidthHeight = width+height      # real size of the input image
    scale = decimal.Decimal(optiWidthHeight)/decimal.Decimal(realWidthHeight)      # ratio for scaling

# save current context
    pdb.gimp_context_push
    
# fix all required context settings 
    pdb.gimp_context_set_antialias(True)
    pdb.gimp_context_set_feather(False)
    pdb.gimp_context_set_feather_radius(0, 0)
    pdb.gimp_context_set_sample_merged(False) 
    pdb.gimp_context_set_sample_criterion(SELECT_CRITERION_COMPOSITE) 
    pdb.gimp_context_set_sample_threshold(0.00)
    pdb.gimp_context_set_sample_transparent(False)       # beware, if TRUE may change the area selected!
    pdb.gimp_context_set_interpolation(INTERPOLATION_LANCZOS)
    
#### START OPERATIONS +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
    
    newwidth=int(width*scale)
    newheight=int(height*scale)
    #pdb.gimp_message("newwidth="+str(newwidth))
    newImage = pdb.gimp_image_new(newwidth, newheight, type)                    
    newBG = pdb.gimp_layer_new_from_drawable(drawable, newImage)     # background layer
    newBG.mode = NORMAL_MODE
    newBG.name = "Background Image Layer"
    newBG.opacity = 100.0 
    newImage.add_layer(newBG, 0)
    #pdb.gimp_desaturate_full(newBG, 1)
    pdb.gimp_context_set_interpolation(INTERPOLATION_LANCZOS)
    pdb.gimp_layer_scale(newBG, newwidth, newheight, False)
 
    alpha=pdb.gimp_drawable_has_alpha(newBG)
    if (alpha==False):
        pdb.gimp_layer_add_alpha (newBG)    #  RGBA
        
    #pdb.gimp_display_new(newImage)    
        
# smooth
# ======
    smooth_layer=pdb.gimp_layer_new_from_drawable(newBG, newImage)
    smooth_layer.name="Smoothing layer"
    newImage.add_layer(smooth_layer, 0)
    pdb.gimp_layer_set_opacity(smooth_layer, 100.0)
    pdb.gimp_layer_set_mode(smooth_layer, SOFTLIGHT_MODE)
    
    smooth_layer=newImage.active_layer
    pdb.plug_in_sel_gauss(newImage, smooth_layer, 25, 50)
    # pls note: smoothing values are very high to better create colour-areas

    smooth_layer=newImage.active_layer

# create quantized layer
# ======================

    quantized_layer=pdb.gimp_layer_new_from_drawable(smooth_layer, newImage) 
    pdb.gimp_item_set_name(quantized_layer, "Quantized Layer "+str(nr_of_colours))  
    newImage.add_layer(quantized_layer, 0)
    pdb.gimp_layer_set_opacity(quantized_layer, 100.0)
    pdb.gimp_layer_set_mode(quantized_layer,NORMAL_MODE)

# quantize the colour_areas using mode indexed
# ============================================

# change image mode to indexed (this produces a cartoonized image, 4...64 colors)
    
    pdb.gimp_image_convert_indexed(newImage, 0, 0, nr_of_colours, 0, 0, "ignored")
    is_indexed=pdb.gimp_drawable_is_indexed(quantized_layer) 
    if is_indexed==1:
        hidden_message="OK, converted"
    else:    
        pdb.gimp_message("ERROR! image not converted, exit")
        return    

# get the number of pixels in the colormap (which exists for indexed images)
# ==========================================================================
    (count, colors)=pdb.gimp_image_get_colormap(newImage)
    # the count is nr_of_colors * 3, example 8*3=24
    nr_areas=(count/3) 
    ##pdb.gimp_message("nr of colors in the palette: "+str(nr_areas)) 
  
    temp_palette=pdb.gimp_palette_new(pdb.gimp_image_get_name(newImage))
    
    index=0
    while index<count:
# Compute the color
        color= (colors[index], colors[index+1], colors[index+2])
# Add to the filter temporary palette
        pdb.gimp_palette_add_entry(temp_palette, str(index/3), color)
        index+=3 
# Endwhile index<count
  
# the temporary palette has been created            
# ======================================

# re-set image to RGB to proceed
# ==============================
    pdb.gimp_image_convert_rgb(newImage)
    alpha=pdb.gimp_drawable_has_alpha(quantized_layer) 
    if alpha ==0:
        pdb.gimp_layer_add_alpha(quantized_layer)     #  RGBA

    (mean, stddev, median, pixels, count, percent)=pdb.gimp_histogram(quantized_layer, 0, 0, 255)
    avgpixels = pixels / nr_areas   # average number of pixels per colour area
    bigareapixels = pixels / 4      # define only a few areas as big, likely just one
    midareapixels = pixels / 16     # define some medium colour areas

#   loop to fill the N areas for colors with wood patterns            <<<<<<<<<<< for all N areas >>>>>>>>>>>
#   ======================================================

    pdb.gimp_image_undo_freeze(newImage)                      # save time and space
    
    if fillType == 0 :   # patterns
        (LTEXT, LFORM)=load_patterns(namesFilter)
    if fillType == 1 :   # gradients
        (LTEXT, LFORM)=load_gradients(namesFilter)
    
    #pdb.gimp_display_new(newImage)
    #return

    areaNR=0
    area_color=[]   # empty list
    for areaNR in range(0,nr_areas):
        one_color=pdb.gimp_palette_entry_get_color(temp_palette, areaNR)
        if selMethod == 1 : # best fit in luma
            best_texture=choose_appropriate_texture(one_color, LFORM, LTEXT)
        else :    # use random texture
            rn = random.randrange (0, len(LTEXT))
            best_texture = LTEXT[rn]     
        pdb.gimp_image_select_color(newImage, CHANNEL_OP_REPLACE, quantized_layer, one_color)
        (there_is_a_selection, x1, y1, x2, y2) = pdb.gimp_selection_bounds(newImage)
        if there_is_a_selection==1:
            (mean, stddev, median, pixels, count, percent)=pdb.gimp_histogram(quantized_layer, 0, 0, 255)
            #gimp.message ("pixels = "+str(pixels))
            
            azimuth_value = 135.0
            if pixels >= bigareapixels :
                width_value = widthheight/200   
                filter_mode = 1 # carve 
            elif pixels >= midareapixels :
                width_value = widthheight/400
                filter_mode = 0 # chisel
            else :  # snall areas
                width_value = widthheight/800
                filter_mode = 0 # chisel
            elev_value = 30     # after tests it looks as the best setting
            depth_value = 20    # after tests it looks as the best setting  
            #pdb.gimp_message("area nr. "+str(areaNR)+" there_is_a_selection")
            color_area=pdb.gimp_image_get_selection(newImage)
            this_channel=pdb.gimp_selection_save(newImage)
            
            if fillType == 0 :   # patterns
                pdb.gimp_context_set_pattern (best_texture)
                pdb.gimp_edit_fill (quantized_layer, PATTERN_FILL)
            if fillType == 1 :   # gradients
                pdb.gimp_context_set_gradient (best_texture)
                (flag, x1, y1, x2, y2) = pdb.gimp_selection_bounds (newImage)
                centrex = (x1+x2)/2
                centrey = (y1+y2)/2
                gtype = random.randrange (1,10) # omit simple linear, not good when start from centre
                pdb.gimp_edit_blend (quantized_layer, CUSTOM_MODE, NORMAL_MODE, gtype, 100.0,
                             0, REPEAT_NONE, False, False, 1, 0, False, centrex, centrey, x2, y2)
                pdb.plug_in_gauss_iir2 (newImage, quantized_layer, newImage.width/200, newImage.height/200)
            
            pdb.script_fu_chisel(newImage, quantized_layer,
                             width_value,             #bevel width
                             10.0*ccRoundness,        #bevel softness,
                             ccRoundness,             #bevel roundness (0.1-1.0),
                             ccRoundness,             #bevel power, #test
                             azimuth_value,           #azimuth  
                             elev_value,              #elevation
                             depth_value,             #depth,  
                             filter_mode,             #mode: 0=chisel,1=carve,
                             0,                       #location,
                             0.0,                     #post-effect blur,
                             False)                   #keep)
            
            #pdb.gimp_display_new(newImage)
            #return
            
            LA = newImage.layers
            #il = len(LA)
            #for i in range(0, il):          
            #    pdb.gimp_message("layer "+str(i)+" has name "+LA[i].name)

            bevel=LA[0]    # should be: ....bevel
            quantized_layer=pdb.gimp_image_merge_down(newImage, bevel, 2)  # should merge ...bevel into quantized
            pdb.gimp_item_set_name(quantized_layer, "Bevel merged (colours="+str(int(nr_of_colours))+")")

            #pdb.gimp_display_new(newImage)
            #return
        
           
    ##pdb.gimp_message("end loop")
    
    pdb.gimp_selection_none(newImage)
    pdb.gimp_layer_set_opacity(quantized_layer, 100.0)
    pdb.gimp_layer_set_mode(quantized_layer, GRAIN_MERGE_MODE)
    
    visible_layer=pdb.gimp_layer_new_from_visible(newImage, newImage, "Now visible Texture: "+namesFilter)
    newImage.add_layer(visible_layer, 0)
    visible_layer.opacity = 45.0
    pdb.gimp_layer_set_mode(visible_layer,SOFTLIGHT_MODE)
    pdb.gimp_item_set_visible (smooth_layer, False) # not to have a double effect
    pdb.gimp_item_set_visible (newBG, False) # not to have a double effect

    if BGonTop == True :
        pdb.gimp_image_raise_item_to_top (newImage, newBG)
        pdb.gimp_item_set_visible (newBG, True) # restore visibility
        
        dupBG=pdb.gimp_layer_new_from_drawable(newBG, newImage)
        dupBG.name="Duplicated BG layer"
        newImage.add_layer(dupBG, 0)
        pdb.gimp_layer_set_mode (dupBG, HARDLIGHT_MODE)
        pdb.gimp_layer_set_opacity(dupBG, 40.0)
        
       # decide which Mode and Opacity depending on the MeanValue of ex-BG layer     
        (meanValue, stdDev, median, pix, cnt, perc) = pdb.gimp_histogram (newBG, HISTOGRAM_VALUE, 0 ,255)
       #gimp.message ("meanValue = "+str(meanValue))
        if meanValue < 96 :    # dark  
            pdb.gimp_layer_set_mode (newBG, COLOR_MODE)
            pdb.gimp_layer_set_opacity (newBG, 95.0)
        elif meanValue > 160 :    # light  
            pdb.gimp_layer_set_mode (newBG, MULTIPLY_MODE)
            pdb.gimp_layer_set_opacity (newBG, 65.0)
        else :  # average      
            pdb.gimp_layer_set_mode (newBG, HARDLIGHT_MODE)
            pdb.gimp_layer_set_opacity (newBG, 55.0)
            
    # if both Post-processings = none, you may set visible only the wood layer by default     
    #if pp_artistic==0 and pp_colors==0 and BGonTop == False:
    #    pdb.gimp_item_set_visible (visible_layer, False)

#   finish the script

    pdb.gimp_palette_delete(temp_palette)
    
    pdb.gimp_context_pop
    
    if (doFlatten == True) :
        pdb.gimp_image_flatten(newImage)
        flattenedLayer = pdb.gimp_image_get_active_layer (newImage)
        pdb.gimp_item_set_name(flattenedLayer,
                               "Texturized (colours="+str(int(nr_of_colours))+"; texture="+namesFilter+")")
  
    pdb.gimp_display_new(newImage)

    return

# This is the plugin registration function  
register(
    "TexturizeByColourAreaSE",
    "Create a quantized image with chosen texture effects",
    "Transform an RGB image into an indexed image and apply a chosen texture effect on each colour area",
    "Diego", 
    "Diego Nassetti ", 
    "May 2016",
    "TexturizeByColourAreaSE",
    "",
    [
      (PF_IMAGE, "image", "Input image", None),
      (PF_DRAWABLE, "drawable", "Input drawable", None),
      (PF_ADJUSTMENT, "size", "Width+Height of New Image", 3300, (1200, 9600, 300)),
      (PF_ADJUSTMENT, "nuances", "How many Nuances", 11, (4, 64, 1)),
      (PF_STRING, "filter", "Filter for names:", "MyWood"),
      (PF_OPTION, "filltype", "Type of fillers", 0, ("Patterns",
                                                     "Gradients",
                                                     )),
      (PF_OPTION, "selmethod", "Fillers selection method", 1, ("Random",
                                                               "Best Luma",
                                                               )),
      (PF_ADJUSTMENT, "roundness", "Chisel/Carve roundness", 0.5, (0.1, 1.0, 0.1)),
      (PF_TOGGLE, "bgontop", "Move BG on top?", True),
      (PF_TOGGLE, "flatten", "Flatten Image?", False),
    ],
    [],
    TexturizeByColourAreaSE,
    menu="<Image>/Diego/Quantize",
    )

main()
